@@ -25,6 +25,7 @@
 
 #include "dnsdist-carbon.hh"
 #include "dnsdist.hh"
+#include "dnsdist-backoff.hh"
 
 #ifndef DISABLE_CARBON
 #include "dolog.hh"
@@ -51,21 +52,24 @@ static bool doOneCarbonExport(const Carbon::Endpoint& endpoint)
 
     const time_t now = time(nullptr);
 
-    for (const auto& e : g_stats.entries) {
-      str << namespace_name << "." << hostname << "." << instance_name << "." << e.first << ' ';
-      if (const auto& val = boost::get<pdns::stat_t*>(&e.second)) {
-        str << (*val)->load();
+    {
+      auto entries = g_stats.entries.read_lock();
+      for (const auto& entry : *entries) {
+        str << namespace_name << "." << hostname << "." << instance_name << "." << entry.d_name << ' ';
+        if (const auto& val = boost::get<pdns::stat_t*>(&entry.d_value)) {
+          str << (*val)->load();
+        }
+        else if (const auto& adval = boost::get<pdns::stat_t_trait<double>*>(&entry.d_value)) {
+          str << (*adval)->load();
+        }
+        else if (const auto& dval = boost::get<double*>(&entry.d_value)) {
+          str << **dval;
+        }
+        else if (const auto& func = boost::get<DNSDistStats::statfunction_t>(&entry.d_value)) {
+          str << (*func)(entry.d_name);
+        }
+        str << ' ' << now << "\r\n";
       }
-      else if (const auto& adval = boost::get<pdns::stat_t_trait<double>*>(&e.second)) {
-        str << (*adval)->load();
-      }
-      else if (const auto& dval = boost::get<double*>(&e.second)) {
-        str << **dval;
-      }
-      else if (const auto& func = boost::get<DNSDistStats::statfunction_t>(&e.second)) {
-        str << (*func)(e.first);
-      }
-      str << ' ' << now << "\r\n";
     }
 
     auto states = g_dstates.getLocal();
@@ -269,6 +273,8 @@ static void carbonHandler(Carbon::Endpoint&& endpoint)
 {
   setThreadName("dnsdist/carbon");
   const auto intervalUSec = endpoint.interval * 1000 * 1000;
+  /* maximum interval between two attempts is 10 minutes */
+  const ExponentialBackOffTimer backOffTimer(10 * 60);
 
   try {
     uint8_t consecutiveFailures = 0;
@@ -287,16 +293,7 @@ static void carbonHandler(Carbon::Endpoint&& endpoint)
         consecutiveFailures = 0;
       }
       else {
-        /* maximum interval between two attempts is 10 minutes */
-        const time_t maxBackOff = 10 * 60;
-        time_t backOff = 1;
-        double backOffTmp = std::pow(2.0, static_cast<double>(consecutiveFailures));
-        if (backOffTmp != HUGE_VAL && static_cast<uint64_t>(backOffTmp) <= static_cast<uint64_t>(std::numeric_limits<time_t>::max())) {
-          backOff = static_cast<time_t>(backOffTmp);
-          if (backOff > maxBackOff) {
-            backOff = maxBackOff;
-          }
-        }
+        const auto backOff = backOffTimer.get(consecutiveFailures);
         if (consecutiveFailures < std::numeric_limits<decltype(consecutiveFailures)>::max()) {
           consecutiveFailures++;
         }
