@@ -30,15 +30,24 @@ std::vector<std::string> KeyValueLookupKeySourceIP::getKeys(const ComboAddress&
   std::vector<std::string> result;
   ComboAddress truncated(addr);
 
+  std::string key;
   if (truncated.isIPv4()) {
     truncated.truncate(d_v4Mask);
-    result.emplace_back(reinterpret_cast<const char*>(&truncated.sin4.sin_addr.s_addr), sizeof(truncated.sin4.sin_addr.s_addr));
+    key.reserve(sizeof(truncated.sin4.sin_addr.s_addr) + (d_includePort ? sizeof(truncated.sin4.sin_port) : 0));
+    key.append(reinterpret_cast<const char*>(&truncated.sin4.sin_addr.s_addr), sizeof(truncated.sin4.sin_addr.s_addr));
   }
   else if (truncated.isIPv6()) {
     truncated.truncate(d_v6Mask);
-    result.emplace_back(reinterpret_cast<const char*>(&truncated.sin6.sin6_addr.s6_addr), sizeof(truncated.sin6.sin6_addr.s6_addr));
+    key.reserve(sizeof(truncated.sin6.sin6_addr.s6_addr) + (d_includePort ? sizeof(truncated.sin4.sin_port) : 0));
+    key.append(reinterpret_cast<const char*>(&truncated.sin6.sin6_addr.s6_addr), sizeof(truncated.sin6.sin6_addr.s6_addr));
   }
 
+  if (d_includePort) {
+    key.append(reinterpret_cast<const char*>(&truncated.sin4.sin_port), sizeof(truncated.sin4.sin_port));
+  }
+
+  result.push_back(std::move(key));
+
   return result;
 }
 
@@ -77,9 +86,8 @@ bool LMDBKVStore::getValue(const std::string& key, std::string& value)
 {
   try {
     auto transaction = d_env.getROTransaction();
-    auto dbi = transaction->openDB(d_dbName, 0);
     MDBOutVal result;
-    int rc = transaction->get(dbi, MDBInVal(key), result);
+    int rc = transaction->get(d_dbi, MDBInVal(key), result);
     if (rc == 0) {
       value = result.get<std::string>();
       return true;
@@ -98,9 +106,8 @@ bool LMDBKVStore::keyExists(const std::string& key)
 {
   try {
     auto transaction = d_env.getROTransaction();
-    auto dbi = transaction->openDB(d_dbName, 0);
     MDBOutVal result;
-    int rc = transaction->get(dbi, MDBInVal(key), result);
+    int rc = transaction->get(d_dbi, MDBInVal(key), result);
     if (rc == 0) {
       return true;
     }
@@ -114,6 +121,54 @@ bool LMDBKVStore::keyExists(const std::string& key)
   return false;
 }
 
+bool LMDBKVStore::getRangeValue(const std::string& key, std::string& value)
+{
+  try {
+    auto transaction = d_env.getROTransaction();
+    auto cursor = transaction->getROCursor(d_dbi);
+    MDBOutVal actualKey;
+    MDBOutVal result;
+    // for range-based lookups, we expect the data in LMDB
+    // to be stored with the last value of the range as key
+    // and the first value of the range as data, sometimes
+    // followed by any other content we don't care about
+    // range-based lookups are mostly useful for network ranges,
+    // for which we expect addresses to be stored in network byte
+    // order
+
+    // retrieve the first key greater or equal to our key
+    int rc = cursor.lower_bound(MDBInVal(key), actualKey, result);
+
+    if (rc == 0) {
+      auto last = actualKey.get<std::string>();
+      if (last.size() != key.size() || key > last) {
+        return false;
+      }
+
+      value = result.get<std::string>();
+      if (value.size() < key.size()) {
+        return false;
+      }
+
+      // take the first part of the data, which should be
+      // the first address of the range
+      auto first = value.substr(0, key.size());
+      if (first.size() != key.size() || key < first) {
+        return false;
+      }
+
+      return true;
+    }
+    else if (rc == MDB_NOTFOUND) {
+      return false;
+    }
+  }
+  catch(const std::exception& e) {
+    vinfolog("Error while looking up a range from LMDB file '%s', database '%s': %s", d_fname, d_dbName, e.what());
+  }
+  return false;
+}
+
 #endif /* HAVE_LMDB */
 
 #ifdef HAVE_CDB
@@ -135,10 +190,9 @@ CDBKVStore::~CDBKVStore() {
 
 bool CDBKVStore::reload(const struct stat& st)
 {
-  auto newCDB = std::unique_ptr<CDB>(new CDB(d_fname));
+  auto newCDB = std::make_unique<CDB>(d_fname);
   {
-    WriteLock wl(&d_lock);
-    d_cdb = std::move(newCDB);
+    *(d_cdb.write_lock()) = std::move(newCDB);
   }
   d_mtime = st.st_mtime;
   return true;
@@ -192,8 +246,8 @@ bool CDBKVStore::getValue(const std::string& key, std::string& value)
     }
 
     {
-      ReadLock rl(&d_lock);
-      if (d_cdb && d_cdb->findOne(key, value)) {
+      auto cdb = d_cdb.read_lock();
+      if (*cdb && (*cdb)->findOne(key, value)) {
         return true;
       }
     }
@@ -214,12 +268,12 @@ bool CDBKVStore::keyExists(const std::string& key)
     }
 
     {
-      ReadLock rl(&d_lock);
-      if (!d_cdb) {
+      auto cdb = d_cdb.read_lock();
+      if (!*cdb) {
         return false;
       }
 
-      return d_cdb->keyExists(key);
+      return (*cdb)->keyExists(key);
     }
   }
   catch(const std::exception& e) {
