@@ -29,6 +29,7 @@
 #include "dnsdist-nghttp2.hh"
 #include "dnsdist-tcp.hh"
 #include "dnsdist-tcp-downstream.hh"
+#include "dnsdist-downstream-connection.hh"
 
 #include "dolog.hh"
 #include "iputils.hh"
@@ -134,7 +135,20 @@ void DoHConnectionToBackend::handleResponse(PendingRequest&& request)
   struct timeval now;
   gettimeofday(&now, nullptr);
   try {
-    request.d_sender->handleResponse(now, TCPResponse(std::move(request.d_buffer), std::move(request.d_query.d_idstate), shared_from_this()));
+    if (!d_healthCheckQuery) {
+      const double udiff = request.d_query.d_idstate.queryRealTime.udiff();
+      d_ds->updateTCPLatency(udiff);
+      if (request.d_buffer.size() >= sizeof(dnsheader)) {
+        dnsheader dh;
+        memcpy(&dh, request.d_buffer.data(), sizeof(dh));
+        d_ds->reportResponse(dh.rcode);
+      }
+      else {
+        d_ds->reportTimeoutOrError();
+      }
+    }
+
+    request.d_sender->handleResponse(now, TCPResponse(std::move(request.d_buffer), std::move(request.d_query.d_idstate), shared_from_this(), d_ds));
   }
   catch (const std::exception& e) {
     vinfolog("Got exception while handling response for cross-protocol DoH: %s", e.what());
@@ -144,6 +158,10 @@ void DoHConnectionToBackend::handleResponse(PendingRequest&& request)
 void DoHConnectionToBackend::handleResponseError(PendingRequest&& request, const struct timeval& now)
 {
   try {
+    if (!d_healthCheckQuery) {
+      d_ds->reportTimeoutOrError();
+    }
+
     request.d_sender->notifyIOError(std::move(request.d_query.d_idstate), now);
   }
   catch (const std::exception& e) {
@@ -191,7 +209,7 @@ bool DoHConnectionToBackend::reachedMaxStreamID() const
 
 bool DoHConnectionToBackend::reachedMaxConcurrentQueries() const
 {
-  //cerr<<"Got "<<getConcurrentStreamsCount()<<" concurrent streams, max is "<<nghttp2_session_get_remote_settings(d_session.get(), NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS)<<endl;
+  // cerr<<"Got "<<getConcurrentStreamsCount()<<" concurrent streams, max is "<<nghttp2_session_get_remote_settings(d_session.get(), NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS)<<endl;
   if (nghttp2_session_get_remote_settings(d_session.get(), NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS) <= getConcurrentStreamsCount()) {
     return true;
   }
@@ -246,7 +264,7 @@ void DoHConnectionToBackend::queueQuery(std::shared_ptr<TCPQuerySender>& sender,
 {
   auto payloadSize = std::to_string(query.d_buffer.size());
 
-  bool addXForwarded = d_ds->d_addXForwardedHeaders;
+  bool addXForwarded = d_ds->d_config.d_addXForwardedHeaders;
 
   /* We use nghttp2_nv_flag.NGHTTP2_NV_FLAG_NO_COPY_NAME and nghttp2_nv_flag.NGHTTP2_NV_FLAG_NO_COPY_VALUE
      to avoid a copy and lowercasing but we need to make sure that the data will outlive the request (nghttp2_on_frame_send_callback called), and that it is already lowercased. */
@@ -259,8 +277,8 @@ void DoHConnectionToBackend::queueQuery(std::shared_ptr<TCPQuerySender>& sender,
   /* Pseudo-headers need to come first (rfc7540 8.1.2.1) */
   addStaticHeader(headers, "method-name", "method-value");
   addStaticHeader(headers, "scheme-name", "scheme-value");
-  addDynamicHeader(headers, "authority-name", d_ds->d_tlsSubjectName);
-  addDynamicHeader(headers, "path-name", d_ds->d_dohPath);
+  addDynamicHeader(headers, "authority-name", d_ds->d_config.d_tlsSubjectName);
+  addDynamicHeader(headers, "path-name", d_ds->d_config.d_dohPath);
   addStaticHeader(headers, "accept-name", "accept-value");
   addStaticHeader(headers, "content-type-name", "content-type-value");
   addStaticHeader(headers, "user-agent-name", "user-agent-value");
@@ -460,14 +478,16 @@ void DoHConnectionToBackend::stopIO()
 {
   d_ioState->reset();
 
-  auto shared = std::dynamic_pointer_cast<DoHConnectionToBackend>(shared_from_this());
-  if (!willBeReusable(false)) {
-    /* remove ourselves from the connection cache, this might mean that our
-       reference count drops to zero after that, so we need to be careful */
-    t_downstreamDoHConnectionsManager.removeDownstreamConnection(shared);
-  }
-  else {
-    t_downstreamDoHConnectionsManager.moveToIdle(shared);
+  if (isIdle()) {
+    auto shared = std::dynamic_pointer_cast<DoHConnectionToBackend>(shared_from_this());
+    if (!willBeReusable(false)) {
+      /* remove ourselves from the connection cache, this might mean that our
+         reference count drops to zero after that, so we need to be careful */
+      t_downstreamDoHConnectionsManager.removeDownstreamConnection(shared);
+    }
+    else {
+      t_downstreamDoHConnectionsManager.moveToIdle(shared);
+    }
   }
 }
 
@@ -716,7 +736,7 @@ int DoHConnectionToBackend::on_stream_close_callback(nghttp2_session* session, i
   conn->d_currentStreams.erase(stream->first);
 
   // cerr<<"Query has "<<request.d_query.d_downstreamFailures<<" failures, backend limit is "<<conn->d_ds->d_retries<<endl;
-  if (request.d_query.d_downstreamFailures < conn->d_ds->d_retries) {
+  if (request.d_query.d_downstreamFailures < conn->d_ds->d_config.d_retries) {
     // cerr<<"in "<<__PRETTY_FUNCTION__<<", looking for a connection to send a query of size "<<request.d_query.d_buffer.size()<<endl;
     ++request.d_query.d_downstreamFailures;
     auto downstream = t_downstreamDoHConnectionsManager.getConnectionToDownstream(conn->d_mplexer, conn->d_ds, now, std::string(conn->d_proxyProtocolPayload));
@@ -726,9 +746,9 @@ int DoHConnectionToBackend::on_stream_close_callback(nghttp2_session* session, i
     conn->handleResponseError(std::move(request), now);
   }
 
-  //cerr<<"we now have "<<conn->getConcurrentStreamsCount()<<" concurrent connections"<<endl;
+  // cerr<<"we now have "<<conn->getConcurrentStreamsCount()<<" concurrent connections"<<endl;
   if (conn->isIdle()) {
-    //cerr<<"stopping IO"<<endl;
+    // cerr<<"stopping IO"<<endl;
     conn->stopIO();
     conn->watchForRemoteHostClosingConnection();
   }
@@ -742,9 +762,9 @@ int DoHConnectionToBackend::on_header_callback(nghttp2_session* session, const n
 
   const std::string status(":status");
   if (frame->hd.type == NGHTTP2_HEADERS && frame->headers.cat == NGHTTP2_HCAT_RESPONSE) {
-    //cerr<<"got header for "<<frame->hd.stream_id<<":"<<endl;
-    //cerr<<"- "<<std::string(reinterpret_cast<const char*>(name), namelen)<<endl;
-    //cerr<<"- "<<std::string(reinterpret_cast<const char*>(value), valuelen)<<endl;
+    // cerr<<"got header for "<<frame->hd.stream_id<<":"<<endl;
+    // cerr<<"- "<<std::string(reinterpret_cast<const char*>(name), namelen)<<endl;
+    // cerr<<"- "<<std::string(reinterpret_cast<const char*>(value), valuelen)<<endl;
     if (namelen == status.size() && memcmp(status.data(), name, status.size()) == 0) {
       auto stream = conn->d_currentStreams.find(frame->hd.stream_id);
       if (stream == conn->d_currentStreams.end()) {
@@ -753,7 +773,7 @@ int DoHConnectionToBackend::on_header_callback(nghttp2_session* session, const n
         return NGHTTP2_ERR_CALLBACK_FAILURE;
       }
       try {
-        stream->second.d_responseCode = pdns_stou(std::string(reinterpret_cast<const char*>(value), valuelen));
+        pdns::checked_stoi_into(stream->second.d_responseCode, std::string(reinterpret_cast<const char*>(value), valuelen));
       }
       catch (...) {
         vinfolog("Error parsing the status header for stream ID %d", frame->hd.stream_id);
@@ -889,7 +909,7 @@ static void dohClientThread(int crossProtocolPipeFD)
     time_t lastTimeoutScan = now.tv_sec;
 
     for (;;) {
-      data.mplexer->run(&now);
+      data.mplexer->run(&now, 1000);
 
       if (now.tv_sec > lastTimeoutScan) {
         lastTimeoutScan = now.tv_sec;
