@@ -303,11 +303,11 @@ static int backendStatTable_handler(netsnmp_mib_handler* handler,
         break;
       case COLUMN_BACKENDLATENCY:
         DNSDistSNMPAgent::setCounter64Value(request,
-                                            server->latencyUsec/1000.0);
+                                            server->getRelevantLatencyUsec() / 1000.0);
         break;
       case COLUMN_BACKENDWEIGHT:
         DNSDistSNMPAgent::setCounter64Value(request,
-                                            server->weight);
+                                            server->d_config.d_weight);
         break;
       case COLUMN_BACKENDOUTSTANDING:
         DNSDistSNMPAgent::setCounter64Value(request,
@@ -331,7 +331,7 @@ static int backendStatTable_handler(netsnmp_mib_handler* handler,
       }
       case COLUMN_BACKENDADDRESS:
       {
-        std::string addr(server->remote.toStringWithPort());
+        std::string addr(server->d_config.remote.toStringWithPort());
         snmp_set_var_typed_value(request->requestvb,
                                  ASN_OCTET_STR,
                                  addr.c_str(),
@@ -341,10 +341,11 @@ static int backendStatTable_handler(netsnmp_mib_handler* handler,
       case COLUMN_BACKENDPOOLS:
       {
         std::string pools;
-        for(auto& p : server->pools) {
-          if(!pools.empty())
+        for (const auto& p : server->d_config.pools) {
+          if (!pools.empty()) {
             pools+=" ";
-          pools+=p;
+          }
+          pools += p;
         }
         snmp_set_var_typed_value(request->requestvb,
                                  ASN_OCTET_STR,
@@ -359,7 +360,7 @@ static int backendStatTable_handler(netsnmp_mib_handler* handler,
         DNSDistSNMPAgent::setCounter64Value(request, server->queries.load());
         break;
       case COLUMN_BACKENDORDER:
-        DNSDistSNMPAgent::setCounter64Value(request, server->order);
+        DNSDistSNMPAgent::setCounter64Value(request, server->d_config.order);
         break;
       default:
         netsnmp_set_request_error(reqinfo,
@@ -374,11 +375,11 @@ static int backendStatTable_handler(netsnmp_mib_handler* handler,
 }
 #endif /* HAVE_NET_SNMP */
 
-bool DNSDistSNMPAgent::sendBackendStatusChangeTrap(const std::shared_ptr<DownstreamState>& dss)
+bool DNSDistSNMPAgent::sendBackendStatusChangeTrap(const DownstreamState& dss)
 {
 #ifdef HAVE_NET_SNMP
-  const string backendAddress = dss->remote.toStringWithPort();
-  const string backendStatus = dss->getStatus();
+  const string backendAddress = dss.d_config.remote.toStringWithPort();
+  const string backendStatus = dss.getStatus();
   netsnmp_variable_list* varList = nullptr;
 
   snmp_varlist_add_variable(&varList,
@@ -393,8 +394,8 @@ bool DNSDistSNMPAgent::sendBackendStatusChangeTrap(const std::shared_ptr<Downstr
                             backendNameOID,
                             OID_LENGTH(backendNameOID),
                             ASN_OCTET_STR,
-                            dss->getName().c_str(),
-                            dss->getName().size());
+                            dss.getName().c_str(),
+                            dss.getName().size());
 
   snmp_varlist_add_variable(&varList,
                             backendAddressOID,
@@ -444,16 +445,16 @@ bool DNSDistSNMPAgent::sendCustomTrap(const std::string& reason)
 bool DNSDistSNMPAgent::sendDNSTrap(const DNSQuestion& dq, const std::string& reason)
 {
 #ifdef HAVE_NET_SNMP
-  std::string local = dq.local->toString();
-  std::string remote = dq.remote->toString();
-  std::string qname = dq.qname->toStringNoDot();
-  const uint32_t socketFamily = dq.remote->isIPv4() ? 1 : 2;
+  std::string local = dq.ids.origDest.toString();
+  std::string remote = dq.ids.origRemote.toString();
+  std::string qname = dq.ids.qname.toStringNoDot();
+  const uint32_t socketFamily = dq.ids.origRemote.isIPv4() ? 1 : 2;
   const uint32_t socketProtocol = dq.overTCP() ? 2 : 1;
   const uint32_t queryType = dq.getHeader()->qr ? 2 : 1;
   const uint32_t querySize = (uint32_t) dq.getData().size();
   const uint32_t queryID = (uint32_t) ntohs(dq.getHeader()->id);
-  const uint32_t qType = (uint32_t) dq.qtype;
-  const uint32_t qClass = (uint32_t) dq.qclass;
+  const uint32_t qType = (uint32_t) dq.ids.qtype;
+  const uint32_t qClass = (uint32_t) dq.ids.qclass;
 
   netsnmp_variable_list* varList = nullptr;
 
