@@ -28,10 +28,13 @@
 #include "dnsname.hh"
 #include "dnsparser.hh"
 #include "dnswriter.hh"
+#include "dolog.hh"
 #include <unistd.h>
 
 bool g_verbose{false};
 bool g_syslog{true};
+bool g_logtimestamps{false};
+std::optional<std::ofstream> g_verboseStream{std::nullopt};
 
 BOOST_AUTO_TEST_SUITE(test_dnscrypt_cc)
 
