@@ -444,6 +444,11 @@ public:
   {
   }
 
+  std::vector<int> getAsyncFDs() override
+  {
+    return {};
+  }
+
   /* unused in that context, don't bother */
   void doHandshake() override
   {
@@ -493,7 +498,7 @@ public:
     return std::make_unique<MockupTLSConnection>(socket);
   }
 
-  std::unique_ptr<TLSConnection> getClientConnection(const std::string& host, int socket, const struct timeval& timeout) override
+  std::unique_ptr<TLSConnection> getClientConnection(const std::string& host, bool hostIsAddr, int socket, const struct timeval& timeout) override
   {
     return std::make_unique<MockupTLSConnection>(socket, true, d_needProxyProtocol);
   }
@@ -594,11 +599,6 @@ public:
     return true;
   }
 
-  const ClientState* getClientState() const override
-  {
-    return nullptr;
-  }
-
   void handleResponse(const struct timeval& now, TCPResponse&& response) override
   {
     if (d_customHandler) {
@@ -630,7 +630,7 @@ public:
   {
   }
 
-  void notifyIOError(IDState&& query, const struct timeval& now) override
+  void notifyIOError(InternalQueryState&& query, const struct timeval& now) override
   {
     d_error = true;
   }
@@ -719,13 +719,13 @@ BOOST_FIXTURE_TEST_CASE(test_SingleQuery, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   auto sender = std::make_shared<MockupQuerySender>();
   sender->d_id = counter;
-  InternalQuery internalQuery(std::move(query), IDState());
+  InternalQuery internalQuery(std::move(query), InternalQueryState());
 
   s_steps = {
     {ExpectedStep::ExpectedRequest::connectToBackend, IOState::Done},
@@ -776,9 +776,9 @@ BOOST_FIXTURE_TEST_CASE(test_ConcurrentQueries, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -803,7 +803,7 @@ BOOST_FIXTURE_TEST_CASE(test_ConcurrentQueries, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -865,9 +865,9 @@ BOOST_FIXTURE_TEST_CASE(test_ConnectionReuse, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -892,7 +892,7 @@ BOOST_FIXTURE_TEST_CASE(test_ConnectionReuse, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -994,9 +994,9 @@ BOOST_FIXTURE_TEST_CASE(test_InvalidDNSAnswer, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   auto sender = std::make_shared<MockupQuerySender>();
   sender->d_id = counter;
@@ -1006,7 +1006,7 @@ BOOST_FIXTURE_TEST_CASE(test_InvalidDNSAnswer, TestFixture)
        while TCP and DoT will first pass it back to the TCP worker thread */
     throw std::runtime_error("Invalid response");
   };
-  InternalQuery internalQuery(std::move(query), IDState());
+  InternalQuery internalQuery(std::move(query), InternalQueryState());
 
   s_steps = {
     {ExpectedStep::ExpectedRequest::connectToBackend, IOState::Done},
@@ -1054,9 +1054,9 @@ BOOST_FIXTURE_TEST_CASE(test_TimeoutWhileWriting, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -1081,7 +1081,7 @@ BOOST_FIXTURE_TEST_CASE(test_TimeoutWhileWriting, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -1116,7 +1116,7 @@ BOOST_FIXTURE_TEST_CASE(test_TimeoutWhileWriting, TestFixture)
   }
 
   struct timeval later = now;
-  later.tv_sec += backend->tcpSendTimeout + 1;
+  later.tv_sec += backend->d_config.tcpSendTimeout + 1;
 
   auto expiredConns = handleH2Timeouts(*s_mplexer, later);
   BOOST_CHECK_EQUAL(expiredConns, 1U);
@@ -1141,9 +1141,9 @@ BOOST_FIXTURE_TEST_CASE(test_TimeoutWhileReading, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -1168,7 +1168,7 @@ BOOST_FIXTURE_TEST_CASE(test_TimeoutWhileReading, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -1204,7 +1204,7 @@ BOOST_FIXTURE_TEST_CASE(test_TimeoutWhileReading, TestFixture)
   }
 
   struct timeval later = now;
-  later.tv_sec += backend->tcpRecvTimeout + 1;
+  later.tv_sec += backend->d_config.tcpRecvTimeout + 1;
 
   auto expiredConns = handleH2Timeouts(*s_mplexer, later);
   BOOST_CHECK_EQUAL(expiredConns, 1U);
@@ -1228,9 +1228,9 @@ BOOST_FIXTURE_TEST_CASE(test_ShortWrite, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -1255,7 +1255,7 @@ BOOST_FIXTURE_TEST_CASE(test_ShortWrite, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -1315,9 +1315,9 @@ BOOST_FIXTURE_TEST_CASE(test_ShortRead, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -1342,7 +1342,7 @@ BOOST_FIXTURE_TEST_CASE(test_ShortRead, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -1409,9 +1409,9 @@ BOOST_FIXTURE_TEST_CASE(test_ConnectionClosedWhileReading, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -1436,7 +1436,7 @@ BOOST_FIXTURE_TEST_CASE(test_ConnectionClosedWhileReading, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -1495,9 +1495,9 @@ BOOST_FIXTURE_TEST_CASE(test_ConnectionClosedWhileWriting, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -1522,7 +1522,7 @@ BOOST_FIXTURE_TEST_CASE(test_ConnectionClosedWhileWriting, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -1589,11 +1589,11 @@ BOOST_FIXTURE_TEST_CASE(test_GoAwayFromServer, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
   /* set the number of reconnection attempts to a low value to not waste time */
-  backend->d_retries = 1;
+  backend->d_config.d_retries = 1;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -1618,7 +1618,7 @@ BOOST_FIXTURE_TEST_CASE(test_GoAwayFromServer, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -1700,9 +1700,9 @@ BOOST_FIXTURE_TEST_CASE(test_HTTP500FromServer, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -1727,7 +1727,7 @@ BOOST_FIXTURE_TEST_CASE(test_HTTP500FromServer, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -1793,9 +1793,9 @@ BOOST_FIXTURE_TEST_CASE(test_WrongStreamID, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -1820,7 +1820,7 @@ BOOST_FIXTURE_TEST_CASE(test_WrongStreamID, TestFixture)
 
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
 
@@ -1868,7 +1868,7 @@ BOOST_FIXTURE_TEST_CASE(test_WrongStreamID, TestFixture)
   }
 
   struct timeval later = now;
-  later.tv_sec += backend->tcpRecvTimeout + 1;
+  later.tv_sec += backend->d_config.tcpRecvTimeout + 1;
 
   auto expiredConns = handleH2Timeouts(*s_mplexer, later);
   BOOST_CHECK_EQUAL(expiredConns, 1U);
@@ -1894,10 +1894,10 @@ BOOST_FIXTURE_TEST_CASE(test_ProxyProtocol, TestFixture)
 
   auto backend = std::make_shared<DownstreamState>(getBackendAddress("42", 53));
   backend->d_tlsCtx = tlsCtx;
-  backend->d_tlsSubjectName = "backend.powerdns.com";
-  backend->d_dohPath = "/dns-query";
-  backend->d_addXForwardedHeaders = true;
-  backend->useProxyProtocol = true;
+  backend->d_config.d_tlsSubjectName = "backend.powerdns.com";
+  backend->d_config.d_dohPath = "/dns-query";
+  backend->d_config.d_addXForwardedHeaders = true;
+  backend->d_config.useProxyProtocol = true;
 
   size_t numberOfQueries = 2;
   std::vector<std::pair<std::shared_ptr<MockupQuerySender>, InternalQuery>> queries;
@@ -1923,7 +1923,7 @@ BOOST_FIXTURE_TEST_CASE(test_ProxyProtocol, TestFixture)
     auto sender = std::make_shared<MockupQuerySender>();
     sender->d_id = counter;
     std::string payload = makeProxyHeader(counter % 2, local, local, {});
-    InternalQuery internalQuery(std::move(query), IDState());
+    InternalQuery internalQuery(std::move(query), InternalQueryState());
     internalQuery.d_proxyProtocolPayload = std::move(payload);
     queries.push_back({std::move(sender), std::move(internalQuery)});
   }
