@@ -13,6 +13,7 @@ BOOST_AUTO_TEST_SUITE(dnsdistrings_cc)
 static void test_ring(size_t maxEntries, size_t numberOfShards, size_t nbLockTries)
 {
   Rings rings(maxEntries, numberOfShards, nbLockTries);
+  rings.init();
   size_t entriesPerShard = maxEntries / numberOfShards;
 
   BOOST_CHECK_EQUAL(rings.getNumberOfShards(), numberOfShards);
@@ -205,7 +206,12 @@ BOOST_AUTO_TEST_CASE(test_Rings_Threaded) {
   dnsdist::Protocol outgoingProtocol = dnsdist::Protocol::DoUDP;
 
   Rings rings(numberOfEntries, numberOfShards, lockAttempts, true);
+  rings.init();
+#if defined(DNSDIST_RINGS_WITH_MACADDRESS)
+  Rings::Query query({requestor, qname, now, dh, size, qtype, protocol, dnsdist::MacAddress(), false});
+#else
   Rings::Query query({requestor, qname, now, dh, size, qtype, protocol});
+#endif
   Rings::Response response({requestor, server, qname, now, dh, latency, size, qtype, outgoingProtocol});
 
   std::atomic<bool> done(false);
