@@ -24,7 +24,7 @@ BOOST_AUTO_TEST_CASE(test_getMultiplexerSilent)
 BOOST_AUTO_TEST_CASE(test_MPlexer)
 {
   for (const auto& entry : FDMultiplexer::getMultiplexerMap()) {
-    auto mplexer = std::unique_ptr<FDMultiplexer>(entry.second());
+    auto mplexer = std::unique_ptr<FDMultiplexer>(entry.second(FDMultiplexer::s_maxevents));
     BOOST_REQUIRE(mplexer != nullptr);
     //cerr<<"Testing multiplexer "<<mplexer->getName()<<endl;
 
@@ -72,7 +72,8 @@ BOOST_AUTO_TEST_CASE(test_MPlexer)
     BOOST_REQUIRE_EQUAL(readyFDs.size(), 1U);
     BOOST_CHECK_EQUAL(readyFDs.at(0), pipes[1]);
 
-    ready = mplexer->run(&now, 100);
+    /* wait until we have at least one descriptor ready */
+    ready = mplexer->run(&now, -1);
     BOOST_CHECK_EQUAL(ready, 1);
     BOOST_CHECK_EQUAL(writeCBCalled, true);
 
@@ -224,7 +225,7 @@ BOOST_AUTO_TEST_CASE(test_MPlexer)
 BOOST_AUTO_TEST_CASE(test_MPlexer_ReadAndWrite)
 {
   for (const auto& entry : FDMultiplexer::getMultiplexerMap()) {
-    auto mplexer = std::unique_ptr<FDMultiplexer>(entry.second());
+    auto mplexer = std::unique_ptr<FDMultiplexer>(entry.second(FDMultiplexer::s_maxevents));
     BOOST_REQUIRE(mplexer != nullptr);
     //cerr<<"Testing multiplexer "<<mplexer->getName()<<" for read AND write"<<endl;
 
